@@ -38,6 +38,8 @@ from .sql import (
     modifier,
     not_,
     null,
+    nullsfirst,
+    nullslast,
     or_,
     outerjoin,
     outparam,
@@ -128,7 +130,7 @@ from .schema import (
 from .inspection import inspect
 from .engine import create_engine, engine_from_config
 
-__version__ = '1.1.6'
+__version__ = '1.2.0'
 
 
 def __go(lcls):
